import numpy as np


def query_func_meta():
    return _dict


def borehole_failmodel(x, theta, p):
    """Given x and theta, return matrix of [row x] times [row theta] of values."""
    ind = np.argmin(np.abs(_dict['fail_array'] - p))
    if p < 0.05:
        ind -= 10
    c = _dict['c_array'][ind]
    f = borehole_model(x, theta)
    wheretoobig = np.where((f / borehole_true(x)) > c)
    f[wheretoobig[0], wheretoobig[1]] = np.nan

    return f


def borehole_failmodel_random(x, theta, p):
    f = borehole_model(x, theta)
    wheretoobig = np.where(
        np.random.choice([0, 1], f.shape, replace=True, p=[1 - p, p]))
    f[wheretoobig[0], wheretoobig[1]] = np.nan
    return f


# def borehole_failmodel_MAR(x, theta, p):
#     """Given x and theta, return matrix of [row x] times [row theta] of values."""
#     f = borehole_model(x, theta)
#     mask = MAR_mask(f, p=p, p_obs=0.5)
#     f[mask] = np.nan
#     return f


def borehole_model(x, theta):
    """Given x and theta, return matrix of [row x] times [row theta] of values."""

    theta = tstd2theta(theta)
    x = xstd2x(x)
    p = x.shape[0]
    n = theta.shape[0]

    theta_stacked = np.repeat(theta, repeats=p, axis=0)
    x_stacked = np.tile(x.astype(float), (n, 1))

    f = borehole_vec(x_stacked, theta_stacked).reshape((n, p))
    return f.T


def borehole_true(x):
    """Given x, return matrix of [row x] times 1 of values."""
    # assume true theta is [0.5]^d
    theta0 = np.atleast_2d(np.array([0.5] * 4))
    f0 = borehole_model(x, theta0)

    return f0


def borehole_vec(x, theta):
    """Given x and theta, return vector of values."""
    (Hu, Ld_Kw, Treff, powparam) = np.split(theta, theta.shape[1], axis=1)
    (rw, Hl) = np.split(x, x.shape[1], axis=1)
    numer = 2 * np.pi * (Hu - Hl)
    denom1 = 2 * Ld_Kw / rw ** 2
    denom2 = Treff

    f = ((numer / ((denom1 + denom2))) * np.exp(powparam * rw)).reshape(-1)
    return f


def tstd2theta(tstd, hard=True):
    """Given standardized theta in [0, 1]^d, return non-standardized theta."""
    if tstd.ndim < 1.5:
        tstd = tstd[:, None].T
    (Treffs, Hus, LdKw, powparams) = np.split(tstd, tstd.shape[1], axis=1)

    Treff = (0.5 - 0.05) * Treffs + 0.05
    Hu = Hus * (1110 - 990) + 990
    if hard:
        Ld_Kw = LdKw * (1680 / 1500 - 1120 / 15000) + 1120 / 15000
    else:
        Ld_Kw = LdKw * (1680 / 9855 - 1120 / 12045) + 1120 / 12045

    powparam = powparams * (0.5 - (- 0.5)) + (-0.5)

    theta = np.hstack((Hu, Ld_Kw, Treff, powparam))
    return theta


def xstd2x(xstd):
    """Given standardized x in [0, 1]^2 x {0, 1}, return non-standardized x."""
    if xstd.ndim < 1.5:
        xstd = xstd[:, None].T
    (rws, Hls) = np.split(xstd, xstd.shape[1], axis=1)

    rw = rws * (np.log(0.5) - np.log(0.05)) + np.log(0.05)
    rw = np.exp(rw)
    Hl = Hls * (820 - 700) + 700

    x = np.hstack((rw, Hl))
    return x


def testborehole():
    np.seterr(all='raise')
    import matplotlib.pyplot as plt
    import scipy.stats.qmc as qmc
    sampler = qmc.LatinHypercube(d=4)
    cs = np.arange(0, 10, 0.01)
    failss = np.zeros((10, *cs.shape))
    for k in np.arange(10):
        x = np.random.uniform(0.1, 0.9, (15, 2))
        theta = sampler.random(1000)
        fails = np.zeros_like(cs)
        for i, c in enumerate(cs):
            f = borehole_failmodel(x, theta, c)
            fails[i] = np.isnan(f).mean()
        failss[k] = fails.copy()
    plt.plot(cs, failss.T)
    plt.show()
    np.seterr()
    return cs, failss


_dict = {
    'function': 'Borehole',
    'xdim': 2,
    'thetadim': 4,
    'c_array': np.arange(0, 10, 0.01),
    'fail_array': np.array(
        [1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 9.999e-01, 9.998e-01, 9.994e-01, 9.988e-01,
         9.979e-01, 9.965e-01, 9.944e-01, 9.913e-01, 9.874e-01, 9.830e-01,
         9.785e-01, 9.734e-01, 9.673e-01, 9.606e-01, 9.539e-01, 9.460e-01,
         9.378e-01, 9.292e-01, 9.197e-01, 9.093e-01, 8.984e-01, 8.871e-01,
         8.757e-01, 8.634e-01, 8.504e-01, 8.373e-01, 8.247e-01, 8.125e-01,
         7.998e-01, 7.875e-01, 7.752e-01, 7.628e-01, 7.512e-01, 7.397e-01,
         7.285e-01, 7.174e-01, 7.062e-01, 6.948e-01, 6.837e-01, 6.735e-01,
         6.634e-01, 6.536e-01, 6.443e-01, 6.354e-01, 6.263e-01, 6.169e-01,
         6.082e-01, 6.004e-01, 5.933e-01, 5.858e-01, 5.785e-01, 5.720e-01,
         5.658e-01, 5.592e-01, 5.524e-01, 5.457e-01, 5.391e-01, 5.323e-01,
         5.255e-01, 5.188e-01, 5.124e-01, 5.061e-01, 5.007e-01, 4.954e-01,
         4.897e-01, 4.846e-01, 4.795e-01, 4.747e-01, 4.699e-01, 4.648e-01,
         4.601e-01, 4.553e-01, 4.506e-01, 4.450e-01, 4.400e-01, 4.351e-01,
         4.305e-01, 4.261e-01, 4.215e-01, 4.171e-01, 4.130e-01, 4.090e-01,
         4.050e-01, 4.011e-01, 3.971e-01, 3.927e-01, 3.886e-01, 3.847e-01,
         3.808e-01, 3.768e-01, 3.729e-01, 3.692e-01, 3.654e-01, 3.614e-01,
         3.577e-01, 3.547e-01, 3.517e-01, 3.486e-01, 3.454e-01, 3.425e-01,
         3.395e-01, 3.367e-01, 3.339e-01, 3.313e-01, 3.286e-01, 3.261e-01,
         3.236e-01, 3.211e-01, 3.188e-01, 3.162e-01, 3.136e-01, 3.109e-01,
         3.081e-01, 3.054e-01, 3.026e-01, 3.000e-01, 2.972e-01, 2.947e-01,
         2.923e-01, 2.898e-01, 2.874e-01, 2.851e-01, 2.828e-01, 2.805e-01,
         2.782e-01, 2.759e-01, 2.736e-01, 2.717e-01, 2.695e-01, 2.675e-01,
         2.653e-01, 2.630e-01, 2.609e-01, 2.588e-01, 2.569e-01, 2.549e-01,
         2.531e-01, 2.514e-01, 2.496e-01, 2.474e-01, 2.455e-01, 2.435e-01,
         2.416e-01, 2.398e-01, 2.381e-01, 2.366e-01, 2.347e-01, 2.329e-01,
         2.312e-01, 2.295e-01, 2.281e-01, 2.265e-01, 2.250e-01, 2.235e-01,
         2.220e-01, 2.205e-01, 2.190e-01, 2.176e-01, 2.162e-01, 2.149e-01,
         2.137e-01, 2.123e-01, 2.110e-01, 2.097e-01, 2.084e-01, 2.070e-01,
         2.055e-01, 2.041e-01, 2.029e-01, 2.014e-01, 2.002e-01, 1.989e-01,
         1.975e-01, 1.964e-01, 1.952e-01, 1.940e-01, 1.929e-01, 1.919e-01,
         1.909e-01, 1.898e-01, 1.886e-01, 1.875e-01, 1.862e-01, 1.852e-01,
         1.839e-01, 1.828e-01, 1.818e-01, 1.809e-01, 1.795e-01, 1.785e-01,
         1.773e-01, 1.763e-01, 1.753e-01, 1.742e-01, 1.732e-01, 1.720e-01,
         1.711e-01, 1.701e-01, 1.688e-01, 1.679e-01, 1.667e-01, 1.656e-01,
         1.647e-01, 1.637e-01, 1.628e-01, 1.618e-01, 1.609e-01, 1.599e-01,
         1.590e-01, 1.581e-01, 1.572e-01, 1.561e-01, 1.552e-01, 1.544e-01,
         1.535e-01, 1.526e-01, 1.517e-01, 1.507e-01, 1.497e-01, 1.488e-01,
         1.478e-01, 1.470e-01, 1.462e-01, 1.452e-01, 1.445e-01, 1.437e-01,
         1.429e-01, 1.422e-01, 1.414e-01, 1.408e-01, 1.400e-01, 1.392e-01,
         1.385e-01, 1.376e-01, 1.367e-01, 1.359e-01, 1.349e-01, 1.342e-01,
         1.334e-01, 1.326e-01, 1.319e-01, 1.311e-01, 1.302e-01, 1.293e-01,
         1.285e-01, 1.277e-01, 1.269e-01, 1.262e-01, 1.255e-01, 1.250e-01,
         1.243e-01, 1.237e-01, 1.231e-01, 1.225e-01, 1.219e-01, 1.214e-01,
         1.207e-01, 1.201e-01, 1.194e-01, 1.186e-01, 1.181e-01, 1.174e-01,
         1.167e-01, 1.161e-01, 1.155e-01, 1.149e-01, 1.143e-01, 1.137e-01,
         1.130e-01, 1.125e-01, 1.119e-01, 1.112e-01, 1.105e-01, 1.097e-01,
         1.091e-01, 1.085e-01, 1.078e-01, 1.072e-01, 1.065e-01, 1.059e-01,
         1.054e-01, 1.049e-01, 1.041e-01, 1.035e-01, 1.029e-01, 1.023e-01,
         1.016e-01, 1.010e-01, 1.004e-01, 9.979e-02, 9.909e-02, 9.846e-02,
         9.789e-02, 9.731e-02, 9.677e-02, 9.625e-02, 9.577e-02, 9.517e-02,
         9.463e-02, 9.419e-02, 9.389e-02, 9.347e-02, 9.301e-02, 9.263e-02,
         9.231e-02, 9.191e-02, 9.148e-02, 9.117e-02, 9.078e-02, 9.036e-02,
         8.989e-02, 8.940e-02, 8.902e-02, 8.861e-02, 8.818e-02, 8.778e-02,
         8.739e-02, 8.708e-02, 8.650e-02, 8.598e-02, 8.563e-02, 8.523e-02,
         8.485e-02, 8.435e-02, 8.395e-02, 8.355e-02, 8.300e-02, 8.261e-02,
         8.211e-02, 8.164e-02, 8.120e-02, 8.083e-02, 8.039e-02, 7.992e-02,
         7.945e-02, 7.901e-02, 7.850e-02, 7.810e-02, 7.768e-02, 7.733e-02,
         7.693e-02, 7.647e-02, 7.607e-02, 7.567e-02, 7.528e-02, 7.489e-02,
         7.448e-02, 7.415e-02, 7.369e-02, 7.329e-02, 7.285e-02, 7.244e-02,
         7.195e-02, 7.158e-02, 7.111e-02, 7.071e-02, 7.017e-02, 6.973e-02,
         6.939e-02, 6.909e-02, 6.873e-02, 6.821e-02, 6.783e-02, 6.748e-02,
         6.715e-02, 6.678e-02, 6.641e-02, 6.600e-02, 6.556e-02, 6.523e-02,
         6.488e-02, 6.460e-02, 6.425e-02, 6.396e-02, 6.362e-02, 6.331e-02,
         6.298e-02, 6.268e-02, 6.239e-02, 6.203e-02, 6.173e-02, 6.144e-02,
         6.112e-02, 6.088e-02, 6.058e-02, 6.035e-02, 6.007e-02, 5.985e-02,
         5.959e-02, 5.923e-02, 5.885e-02, 5.860e-02, 5.823e-02, 5.789e-02,
         5.754e-02, 5.715e-02, 5.685e-02, 5.650e-02, 5.621e-02, 5.589e-02,
         5.560e-02, 5.529e-02, 5.494e-02, 5.459e-02, 5.426e-02, 5.387e-02,
         5.349e-02, 5.319e-02, 5.291e-02, 5.263e-02, 5.233e-02, 5.205e-02,
         5.183e-02, 5.158e-02, 5.131e-02, 5.101e-02, 5.075e-02, 5.049e-02,
         5.021e-02, 4.987e-02, 4.958e-02, 4.931e-02, 4.901e-02, 4.871e-02,
         4.835e-02, 4.799e-02, 4.771e-02, 4.741e-02, 4.713e-02, 4.681e-02,
         4.661e-02, 4.633e-02, 4.606e-02, 4.571e-02, 4.542e-02, 4.513e-02,
         4.487e-02, 4.457e-02, 4.436e-02, 4.405e-02, 4.378e-02, 4.353e-02,
         4.325e-02, 4.303e-02, 4.277e-02, 4.252e-02, 4.226e-02, 4.200e-02,
         4.175e-02, 4.142e-02, 4.112e-02, 4.090e-02, 4.061e-02, 4.036e-02,
         4.008e-02, 3.977e-02, 3.950e-02, 3.926e-02, 3.897e-02, 3.867e-02,
         3.841e-02, 3.813e-02, 3.792e-02, 3.770e-02, 3.752e-02, 3.727e-02,
         3.703e-02, 3.676e-02, 3.649e-02, 3.619e-02, 3.599e-02, 3.582e-02,
         3.557e-02, 3.529e-02, 3.509e-02, 3.487e-02, 3.461e-02, 3.432e-02,
         3.404e-02, 3.380e-02, 3.360e-02, 3.338e-02, 3.312e-02, 3.295e-02,
         3.279e-02, 3.257e-02, 3.238e-02, 3.222e-02, 3.203e-02, 3.189e-02,
         3.174e-02, 3.149e-02, 3.129e-02, 3.117e-02, 3.101e-02, 3.083e-02,
         3.063e-02, 3.049e-02, 3.030e-02, 3.013e-02, 2.995e-02, 2.977e-02,
         2.962e-02, 2.946e-02, 2.933e-02, 2.916e-02, 2.897e-02, 2.889e-02,
         2.867e-02, 2.852e-02, 2.831e-02, 2.816e-02, 2.801e-02, 2.781e-02,
         2.763e-02, 2.745e-02, 2.727e-02, 2.712e-02, 2.696e-02, 2.681e-02,
         2.665e-02, 2.651e-02, 2.639e-02, 2.623e-02, 2.606e-02, 2.592e-02,
         2.575e-02, 2.566e-02, 2.545e-02, 2.525e-02, 2.506e-02, 2.490e-02,
         2.475e-02, 2.462e-02, 2.451e-02, 2.436e-02, 2.425e-02, 2.409e-02,
         2.393e-02, 2.380e-02, 2.355e-02, 2.335e-02, 2.319e-02, 2.309e-02,
         2.293e-02, 2.282e-02, 2.265e-02, 2.253e-02, 2.231e-02, 2.214e-02,
         2.195e-02, 2.182e-02, 2.169e-02, 2.154e-02, 2.136e-02, 2.119e-02,
         2.101e-02, 2.084e-02, 2.076e-02, 2.060e-02, 2.044e-02, 2.027e-02,
         2.011e-02, 1.994e-02, 1.978e-02, 1.961e-02, 1.947e-02, 1.931e-02,
         1.915e-02, 1.892e-02, 1.876e-02, 1.857e-02, 1.843e-02, 1.828e-02,
         1.819e-02, 1.805e-02, 1.788e-02, 1.773e-02, 1.755e-02, 1.737e-02,
         1.724e-02, 1.708e-02, 1.686e-02, 1.670e-02, 1.649e-02, 1.628e-02,
         1.611e-02, 1.591e-02, 1.583e-02, 1.569e-02, 1.556e-02, 1.537e-02,
         1.524e-02, 1.506e-02, 1.491e-02, 1.475e-02, 1.461e-02, 1.439e-02,
         1.418e-02, 1.397e-02, 1.380e-02, 1.362e-02, 1.347e-02, 1.327e-02,
         1.313e-02, 1.301e-02, 1.290e-02, 1.276e-02, 1.261e-02, 1.243e-02,
         1.230e-02, 1.219e-02, 1.201e-02, 1.189e-02, 1.177e-02, 1.166e-02,
         1.155e-02, 1.143e-02, 1.133e-02, 1.123e-02, 1.115e-02, 1.107e-02,
         1.099e-02, 1.087e-02, 1.074e-02, 1.065e-02, 1.059e-02, 1.047e-02,
         1.035e-02, 1.027e-02, 1.013e-02, 1.004e-02, 9.880e-03, 9.793e-03,
         9.660e-03, 9.493e-03, 9.387e-03, 9.253e-03, 9.153e-03, 9.060e-03,
         8.947e-03, 8.847e-03, 8.740e-03, 8.660e-03, 8.593e-03, 8.533e-03,
         8.427e-03, 8.373e-03, 8.267e-03, 8.147e-03, 8.087e-03, 8.053e-03,
         7.960e-03, 7.847e-03, 7.793e-03, 7.693e-03, 7.627e-03, 7.573e-03,
         7.473e-03, 7.427e-03, 7.360e-03, 7.287e-03, 7.213e-03, 7.147e-03,
         7.087e-03, 6.980e-03, 6.920e-03, 6.820e-03, 6.740e-03, 6.653e-03,
         6.580e-03, 6.500e-03, 6.413e-03, 6.333e-03, 6.213e-03, 6.140e-03,
         6.080e-03, 6.033e-03, 5.953e-03, 5.867e-03, 5.800e-03, 5.713e-03,
         5.613e-03, 5.540e-03, 5.460e-03, 5.440e-03, 5.393e-03, 5.320e-03,
         5.273e-03, 5.233e-03, 5.180e-03, 5.100e-03, 5.040e-03, 4.993e-03,
         4.947e-03, 4.867e-03, 4.800e-03, 4.733e-03, 4.647e-03, 4.587e-03,
         4.540e-03, 4.473e-03, 4.433e-03, 4.407e-03, 4.327e-03, 4.287e-03,
         4.260e-03, 4.173e-03, 4.133e-03, 4.107e-03, 4.033e-03, 4.027e-03,
         3.967e-03, 3.913e-03, 3.853e-03, 3.800e-03, 3.753e-03, 3.680e-03,
         3.667e-03, 3.607e-03, 3.547e-03, 3.513e-03, 3.480e-03, 3.440e-03,
         3.387e-03, 3.353e-03, 3.287e-03, 3.227e-03, 3.173e-03, 3.107e-03,
         3.087e-03, 3.060e-03, 3.027e-03, 2.980e-03, 2.953e-03, 2.927e-03,
         2.913e-03, 2.860e-03, 2.813e-03, 2.780e-03, 2.747e-03, 2.700e-03,
         2.667e-03, 2.647e-03, 2.633e-03, 2.600e-03, 2.580e-03, 2.560e-03,
         2.540e-03, 2.520e-03, 2.500e-03, 2.487e-03, 2.460e-03, 2.447e-03,
         2.427e-03, 2.407e-03, 2.380e-03, 2.347e-03, 2.320e-03, 2.307e-03,
         2.287e-03, 2.280e-03, 2.240e-03, 2.227e-03, 2.213e-03, 2.173e-03,
         2.160e-03, 2.140e-03, 2.127e-03, 2.120e-03, 2.100e-03, 2.080e-03,
         2.053e-03, 2.040e-03, 2.013e-03, 1.987e-03, 1.953e-03, 1.927e-03,
         1.927e-03, 1.887e-03, 1.880e-03, 1.860e-03, 1.840e-03, 1.827e-03,
         1.800e-03, 1.780e-03, 1.747e-03, 1.733e-03, 1.693e-03, 1.667e-03,
         1.633e-03, 1.607e-03, 1.593e-03, 1.560e-03, 1.527e-03, 1.507e-03,
         1.453e-03, 1.433e-03, 1.393e-03, 1.373e-03, 1.340e-03, 1.327e-03,
         1.307e-03, 1.287e-03, 1.260e-03, 1.247e-03, 1.220e-03, 1.200e-03,
         1.200e-03, 1.187e-03, 1.187e-03, 1.173e-03, 1.160e-03, 1.140e-03,
         1.113e-03, 1.107e-03, 1.100e-03, 1.087e-03, 1.073e-03, 1.060e-03,
         1.027e-03, 1.020e-03, 9.867e-04, 9.667e-04, 9.667e-04, 9.667e-04,
         9.600e-04, 9.467e-04, 9.467e-04, 9.333e-04, 9.267e-04, 9.200e-04,
         9.000e-04, 8.933e-04, 8.867e-04, 8.800e-04, 8.733e-04, 8.600e-04,
         8.533e-04, 8.467e-04, 8.467e-04, 8.333e-04, 8.333e-04, 8.267e-04,
         8.267e-04, 8.200e-04, 8.200e-04, 8.133e-04, 8.067e-04, 8.067e-04,
         8.000e-04, 8.000e-04, 7.933e-04, 7.867e-04, 7.800e-04, 7.667e-04,
         7.600e-04, 7.467e-04, 7.400e-04, 7.333e-04, 7.333e-04, 7.133e-04,
         7.067e-04, 7.067e-04, 7.000e-04, 6.867e-04, 6.733e-04, 6.667e-04,
         6.400e-04, 6.267e-04, 6.200e-04, 6.067e-04, 5.867e-04, 5.733e-04,
         5.667e-04, 5.467e-04, 5.333e-04, 5.267e-04, 5.200e-04, 5.133e-04,
         5.067e-04, 5.067e-04, 5.067e-04, 5.000e-04, 4.933e-04, 4.933e-04,
         4.867e-04, 4.667e-04, 4.600e-04, 4.600e-04, 4.600e-04, 4.600e-04,
         4.533e-04, 4.467e-04, 4.333e-04, 4.333e-04, 4.333e-04, 4.067e-04,
         4.000e-04, 4.000e-04, 3.933e-04, 3.800e-04, 3.733e-04, 3.667e-04,
         3.600e-04, 3.533e-04, 3.467e-04, 3.200e-04, 3.200e-04, 3.133e-04,
         3.133e-04, 3.133e-04, 3.133e-04, 3.133e-04, 3.133e-04, 3.133e-04,
         3.133e-04, 3.067e-04, 3.000e-04, 3.000e-04, 3.000e-04, 3.000e-04,
         2.933e-04, 2.933e-04, 2.867e-04, 2.867e-04, 2.667e-04, 2.667e-04,
         2.600e-04, 2.467e-04, 2.400e-04, 2.400e-04, 2.400e-04, 2.400e-04,
         2.333e-04, 2.267e-04, 2.133e-04, 2.067e-04, 1.933e-04, 1.933e-04,
         1.800e-04, 1.800e-04, 1.667e-04, 1.600e-04, 1.600e-04, 1.533e-04,
         1.400e-04, 1.400e-04, 1.333e-04, 1.267e-04, 1.200e-04, 1.200e-04,
         1.200e-04, 1.133e-04, 1.067e-04, 1.067e-04, 1.067e-04, 1.000e-04,
         1.000e-04, 8.667e-05, 7.333e-05, 6.667e-05])
}
